/***************************************************************************/
/*                                                                         */
/*  File: VerletModel.h                                                    */
/*  Date: 24-10-07                                                         */
/*                                                                         */
/***************************************************************************/
/*
	Simple class for rendering raw model data which we export and include
	using header files...
*/
/***************************************************************************/
#ifndef VERLETMODEL_H
#define VERLETMODEL_H


#include <d3dx9.h>				// Direct X Header Files

								// Direct X Library Files
#pragma comment(lib, "d3d9.lib")	
#pragma comment(lib, "d3dx9.lib")

// DirectX Device/Render Handles
extern LPDIRECT3D9			g_pD3D;			// main.cpp
extern LPDIRECT3DDEVICE9	g_pD3DDevice;	// main.cpp

#include "DrawModel.h"


 // Some helper containers
struct Constraint
{
	int		index0;
	int		index1;
	float	restLength;
};

struct Point
{
	D3DXVECTOR3 curPos;
	D3DXVECTOR3 oldPos;
	bool		unmovable;
};



class VerletModel
{
public:
	stModelData * m_modelData;

	int			 m_numPoints;
	Point*		 m_points;
	int			 m_numConstraints;
	Constraint*  m_constraints;
	D3DXVECTOR3  m_offset;
	float		 m_scale;


	DrawModel	 m_drawModel;

	VerletModel(stModelData* md, IDirect3DDevice9* pDevice)
	{
		m_modelData			= NULL;
		m_numPoints			= 0;
		m_points			= NULL;
		m_numConstraints	= 0;
		m_constraints		= NULL;
		
		Create( md, pDevice );
	}
	~VerletModel(){Release();}

	void Create(stModelData* md, IDirect3DDevice9* pDevice)
	{
		m_modelData = md;
		
		m_drawModel.Create(md, pDevice);

		m_offset = D3DXVECTOR3(0,12,0);
		m_scale	 = 0.3f;
		m_numPoints = md->numVerts;
		m_points = new Point[m_numPoints];

		m_numConstraints = (m_numPoints * m_numPoints) - m_numPoints;
		m_constraints = new Constraint[m_numConstraints];

		for (int i=0; i<m_numPoints; i++)
		{
			float * data = &md->v[i*3];
			D3DXVECTOR3* pos = (D3DXVECTOR3*)data;
			m_points[i].curPos = (*pos)*m_scale + m_offset;
			m_points[i].oldPos = (*pos)*m_scale + m_offset;
			m_points[i].unmovable = false;
		}
		
		int c = 0;
		for (int i=0; i<m_numPoints; i++)
		{
			for (int k=0; k<m_numPoints; k++)
			{
				if (i!=k)
				{
					Point* p0 = &m_points[i];
					Point* p1 = &m_points[k];

					float len = D3DXVec3Length(&(p0->curPos - p1->curPos));
					m_constraints[c].restLength = len;
					m_constraints[c].index0 = i;
					m_constraints[c].index1 = k;
					c++;
				}
			}
		}

		
		if (c!=m_numConstraints) DBG_HALT;
	}

	void Release()
	{
		delete[] m_points;
		delete[] m_constraints;
	}

	void VerletIntegrate(float dt)
	{
		for (int i=0; i<m_numPoints; i++)
		{
			if (!m_points[i].unmovable)
			{
				D3DXVECTOR3 oldPos = m_points[i].oldPos;
				D3DXVECTOR3 curPos = m_points[i].curPos;
				D3DXVECTOR3 a = D3DXVECTOR3(0,-9.8f,0);

				curPos = 2*curPos - oldPos + a*dt*dt;

				m_points[i].oldPos = m_points[i].curPos;
				m_points[i].curPos = curPos;
			}
		}
	}

	void SatisfyConstraints()
	{
		const int numIterations = 1;

		for (int i=0; i<numIterations; i++)
		{
			for (int k=0; k< m_numConstraints; k++)
			{
				// Constraint 1 (Floor)
				for (int v=0; v<m_numPoints; v++)
				{
					if (m_points[v].curPos.y < 0.0f) m_points[v].curPos.y = 0.0f;
				}

				// Constraint 2 (Links)
				Constraint* c = &m_constraints[k];
				D3DXVECTOR3& p0 = m_points[c->index0].curPos;
				D3DXVECTOR3& p1 = m_points[c->index1].curPos;
				D3DXVECTOR3 delta = p1-p0;
				float len = D3DXVec3Length(&delta);
				float diff = (len - c->restLength) / len;
				p0 += delta*0.5f*diff;
				p1 -= delta*0.5f*diff;


				if (m_points[c->index0].unmovable)
				{
					m_points[c->index0].curPos = m_points[c->index0].oldPos;
				}
				if (m_points[c->index1].unmovable)
				{
					m_points[c->index1].curPos = m_points[c->index1].oldPos;
				}

			}
		}
	}

	void UpdateVerts()
	{
		IDirect3DVertexBuffer* vb = m_drawModel.GetVertexBuffer();

		DrawModel::CUSTOMVERTEX *pVertex	= NULL;
		vb->Lock(0,0, (void**)&pVertex, 0);
		for (int i=0; i<(int)m_numPoints; i++)
		{
			pVertex[i].x = m_points[i].curPos.x;
			pVertex[i].y = m_points[i].curPos.y;
			pVertex[i].z = m_points[i].curPos.z;
		}
		vb->Unlock();
	};

	void Update(float dt)
	{
		VerletIntegrate(dt);
		SatisfyConstraints();
	}


	void Render(IDirect3DDevice9* pDevice)
	{
		UpdateVerts();

		// Draw a wire mesh for our constraints
		for (int i=0; i<m_numConstraints; i++)
		{
			Constraint* c = &m_constraints[i];
			DrawLine3D( pDevice, 
						m_points[c->index0].curPos, 
						m_points[c->index1].curPos,
						0xffff00ff);
		}

		m_drawModel.Render(pDevice);
	}

};






#endif //VERLETMODEL_H



